__all__ = [
    'Settings',
]

import os
from collections import ChainMap
from starlette.config import Config
from starlette.datastructures import Secret, CommaSeparatedStrings
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates
from dataclasses import dataclass, field



WORKING_DIRECTORY = os.getcwd()
ENV_FILE = os.path.abspath('.env')

static = StaticFiles(directory=os.path.join(WORKING_DIRECTORY, 'static'))
jinja = Jinja2Templates(directory=os.path.join(WORKING_DIRECTORY, 'templates'))
config = Config(env_file=ENV_FILE)



@dataclass
class Settings:
    config: Config = config
    static: StaticFiles = static 
    jinja: Jinja2Templates = jinja
        
    def __post_init__(self):
        self.chain = ChainMap({}, {k: Secret(v) for k,v in vars(self.config)['file_values'].items()})
        
    @property
    def cwd(self):
        return os.getcwd()
    
    @property
    def env_file(self):
        return os.path.abspath('.env')

    def update(self, data: dict):
        self.chain.update(data)

    def get(self, key):
        return Secret(self.chain.get(key))
    
    def all_chain(self):
        return self.chain
    
    def initial(self):
        return {k:Secret(v) for k,v in self.chain.maps[-1].items()}    
    


#!/usr/bin/env python
# coding: utf-8
"""Utils module"""


__all__ = [
    'json_encode',
    'is_private',
    'unprivate',
    'remove_accents',
    'today',
    'normalize',
    'age',
    'today',
    'now',
    'is_defined',
    'is_missing',
    'exist',
    'only_digits',
    'no_digits',
    'split_digits_and_letters',
    'zip_longest',
    'semicolon_split',
    'comma_split',
    'list_phrases',
    'process_branded_drug',
    'cammel_case',
    'list_words',
    'group_words',
    'check'
    ]

import re
import itertools
from enum import Enum 
from dataclasses import MISSING
from datetime import timedelta, date, datetime
from decimal import Decimal
import unicodedata 
from typing import Union

today = date.today
now = datetime.now
normalize = lambda string: unicodedata.normalize('NFKD', string)

def age(birthdate: Union[date, str]) -> int:
    return ((today() - birthdate if isinstance(birthdate, date) else date.fromisoformat(birthdate)).days/365).__round__(2)

def remove_accents(string: str) -> str:
    return ''.join([c for c in normalize(string) if not unicodedata.combining(c)])

def is_private(string: str) -> bool:
    return True if string.startswith("_") and not string.endswith("_") else False

def unprivate(string: str) -> str:
    return string[ 1: ] if is_private(string) else string

def json_encode(value):
    if value == None:
        return ''
    elif isinstance(value, bool):
        return value
    elif isinstance(value, (str, int, float)):
        return value
    elif isinstance(value, Decimal):
        return str(value)
    elif isinstance(value, list):
        return [json_encode(i) for i in value]
    elif isinstance(value, dict):
        return {unprivate(k): json_encode(v) for k, v in value.items()}
    elif isinstance(value, Enum):
        return str(value)
    elif isinstance(value, (datetime, date)):
        return value.isoformat()
    elif isinstance(value, timedelta):
        return f'{value.days} {"dias"  if value.days > 1 else "dia"}'
    elif isinstance(value, tuple):
        try:
            return {k: json_encode(v) for (k,v) in value._asdict().items()}
        except:
            return [json_encode(v) for v in value]
    else:
        return str(value)

def is_defined(value):
    return False if value == '<undefined>' else True

def is_missing(value):
    return True if value == '<missing>' else False

def exist(value):
    return False if value in ['', None, '<missing>', '<undefined>'] else True

def semicolon_split(string: str):
    return [x.strip() for x in string.split(';')]

def comma_split(string: str):
    return [x.strip() for x in string.split(',')]

def list_phrases(string: str):
    norm = lambda x: x[0].upper() + x[1:] + '.'
    data = [x.strip() for x in re.split(r'\.\s', string) if exist(x)]
    return [norm(x) for x in data]

def only_digits(value):
    return ''.join(filter(lambda x: x.isdigit() or x in ['.'], value))

def no_digits(value):
    return ''.join(filter(lambda x: True if not x.isdigit() and not x in ['.'] else False, value))

def split_digits_and_letters(value):
    '''Returns a tuple with numbers as index 0 and letters as index 1'''
    return eval(only_digits(value)) if only_digits(value) else None, no_digits(value)


def zip_longest(*args):
    return itertools.zip_longest(*args)

def process_branded_drug(brand: dict):
    presentations = semicolon_split(brand['sets'])
    strengths, quantities = [], []
    for item in presentations:
        data = [x.strip() for x in item.split()]
        strengths.append(data[:-1])
        quantities.append(data[-1])
    return [[semicolon_split(brand['drugs']), *x, brand['recipe'], brand['name'], brand['pharmaceutical']] for x in zip_longest(strengths, quantities)]

def cammel_case(string: str):
    return ''.join([x.strip() for x in remove_accents(string).split()])


def list_words(string: str):
    return [x.strip() for x in string.split(' ')]



def group_words(text: str):
    p = re.compile(r'(?P<word>\b\w+\b)')
    return p.findall(text)


def check(function, iterable):
    gen = (x for x in iterable)
    result = []
    try:
        while True:
            result.append(function(next(gen)))
    except StopIteration:
        pass 
    finally:
        return any(result)


#!/usr/bin/env python
# coding: utf-8
from __future__ import annotations

'''metadoctor.base module'''

__all__ = [
    'BaseModel',
    'BaseField',
    'BaseManager',
    'Manager',
    'RegularField',
    'EnumField', 
    'ReferenceField', 
    'DateTime', 
    'Text', 
    'SemicolonStrToList',
    'String', 
    'Choice',
    'SelectEnum',
    'Date',
    'Title',
    'Number', 
    'Integer',
    'Float',
    'Real', 
    'Boolean',
    'Key',
    'KeyData',
    'Array',
    'ArrayOfKeys',
    'DataField',
    'Instance'
]

import asyncio 
from abc import ABC
from collections import ChainMap, namedtuple
from markupsafe import Markup
import re
import datetime
from abc import ABC
from collections import ChainMap
from decimal import Decimal
from typing import Any
from markupsafe import Markup
import asyncio
from abc import ABC
from functools import cache
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union, NamedTuple, List

from metadoctor import enumeration
from metadoctor import element
from metadoctor import utils
from metadoctor import database as db

# aliases and lambdas
Connection = db.Connection
is_defined = enumeration.Variable.is_defined
exist = enumeration.Variable.exist
Project = enumeration.Project
BaseEnum = enumeration.BaseEnum
Form = element.Form
Fieldset = element.Fieldset
BaseEnum = enumeration.BaseEnum
Variable = enumeration.Variable
is_defined = Variable.is_defined
exist = Variable.exist
process_undefined = lambda x: None if x in [None, Variable.UNDEFINED] else x
Fieldset = element.Fieldset
Input = element.Input
TextArea = element.TextArea
Select = element.Select
Li = element.Li
FormField = element.FormField
Div = element.Div
Option = element.Option
DataList = element.DataList
column_set = lambda x,y,z: f'col-sm-{x} col-md-{y} col-lg-{z}'
only_digits = utils.only_digits

INPUT_STYLE = element.INPUT_STYLE
POSITIONAL_HTML_ELEMENT_ATTRIBUTES = element.POSITIONAL_HTML_ELEMENT_ATTRIBUTES
KEYWORD_VALUE_HTML_ELEMENT_ATTRIBUTES = element.KEYWORD_VALUE_HTML_ELEMENT_ATTRIBUTES
BASE_FIELD_CONFIG_KEYS = 'transform fsize fclass fstyle tag default factory exclude title description options'.split()



class BaseModel(ABC):
    __project = None
    __table = None
    __objname = None
    __singular = None
    __plural = None

    def __init__(self, *args, **kwargs) -> None:
        self._extra_data = dict()
        for base in [*type(self).__bases__, type(self)]:
            if issubclass(base, BaseModel):
                for k, field in base.fields().items():
                    setattr(type(self), k, field) 
        for k in self.fields().keys():
            if k in kwargs.keys():
                setattr(self, k, kwargs[k])
            else:
                setattr(self, k, None)
        
        for k,v in kwargs.items():
            if not k in self.fields().keys():
                self._extra_data[k] = v
                
        self._key = kwargs.get('key') or self._autokey()

        self.__post_init__()
        
        
    def __post_init__(self):
        pass 
    
    def _autokey(self):
        return None
    
    
    def __str__(self) -> str:
        return ', '.join([str(v) for v in self.main_data().values() if exist(v)])

    def __repr__(self) -> str:
        return f'{type(self).__name__}({", ".join([f"{k}={str(v)}" for k, v in self.main_data().items() if exist(v)])})' 
    
    def __hash__(self):
        return hash(tuple(self.main_json().values()))
    
    def __eq__(self, other):
        return repr(self) == repr(other)


    def __repr_html__(self) -> str:
        return Markup(f'<h3>{str(self)}</h3>')

    def __html__(self) -> str:
        return self.__repr_html__()
    
    @classmethod
    @cache
    async def list_all(cls):
        async with getattr(Connection(), cls.project()).ListAll(cls.table()) as data:
            return [cls(**item) for item in data]
        
    @classmethod
    @cache
    async def options(cls, selected=None):
        data = await cls.list_all()
        result = list(sorted(data, key=lambda x: str(x)))
        options = [Option(value=x.key, content=str(x)) for x in result]
        for item in options:
            if item == selected:
                item.selected = True 
        return ''.join([str(x) for x in options])
        
        
    @cache
    async def _exist(self):
        data = await self.list_all()
        result = list(filter(lambda x: x == self, data))
        return None if len(result) == 0 else result[0]
    
    

        
        
    @classmethod
    @cache
    async def by_key(cls, key: str):
        async with getattr(Connection(), cls.project()).Get(cls.table(), key) as result:
            if result:
                return cls(**result)
            return None
        
    @cache
    async def from_db(self):
        async with getattr(Connection, self.project()).Get(self.table(), self.key) as result:
            if result:
                return type(self)(**result)
            return None
        

    async def update(self, **kwargs):
        data = self.json().update(**kwargs)
        async with getattr(Connection(), self.project()).Put(self.table(), data) as result:
            if result:
                return type(self)(**result)
            return None
        
    @cache
    async def get_or_create(self):
        exist, created = await self._exist(), None
        if not exist:
            async with getattr(Connection(), self.project()).Put(self.table(), self.json()) as result:
                if result:
                    created = type(self)(**result)
        return exist, created
    
    
    @classmethod
    @cache
    async def list_all_enum(cls):
        return BaseEnum(f'{cls.__name__}Enum', [(x.key, x) for x in await cls.list_all()])
    
    
    @classmethod
    def ismodel(cls):
        return True 
    
    @classmethod
    def table_project(cls):
        return cls.table(), cls.project()
    
    @staticmethod
    def json_encode(v):
        return utils.json_encode(v)
                
    @classmethod
    def fields(cls):
        return {k:v for (k,v) in vars(cls).items() if isinstance(v, BaseField)}         
    
    @classmethod
    def namedtuple(cls):
        ntuple = namedtuple(f'{cls.table()}NamedTuple', ' '.join(cls.fields().keys()) + ' key')
        ntuple.__model__ = cls
        ntuple.model_instance = lambda x: cls(**x._asdict())
        return ntuple
    
    def astuple(self):
        return self.namedtuple()(**self.data())
    
    def data(self) -> dict:
        result = {k:v.__get__(self) for (k,v) in self.fields().items()}
        if exist(self.key):
            result['key'] = self.key
        return result 
    
    def json(self) -> dict:
        result = {k: self.json_encode(v.unparsed(self)) for k,v in self.fields().items()}
        if exist(self.key):
            result['key'] = self.key
        return result 
    
    def main_data(self) -> str:
        return {k: v.__get__(self) for k, v in self.fields().items() if not is_defined(v.exclude)}
    
    
    def main_json(self) -> str:
        return {k: self.json_encode(v.unparsed(self)) for k, v in self.fields().items() if not is_defined(v.exclude) if v.unparsed(self)}
    
    def extra_data(self):
        return self._extra_data
    
    @classmethod
    def table(cls):
        name = f'_{cls.__name__}__table'
        return cls.__name__ if not hasattr(cls, name) else getattr(cls, name)
            
    @classmethod
    def project(cls):
        name = f'_{cls.__name__}__project'
        return Project.BASE if not hasattr(cls, name) else getattr(cls, name)
    
    
    @classmethod
    def objname(cls):
        name = f'_{cls.__name__}__objname'
        return cls.__name__.lower() if not hasattr(cls, name) else getattr(cls, name)
    
    @classmethod
    def singular(cls):
        name = f'_{cls.__name__}__singular'
        return cls.__name__ if not hasattr(cls, name) else getattr(cls, name)
    
    @classmethod
    def plural(cls):
        name = f'_{cls.__name__}__plural'
        return f'{cls.__name__}s' if not hasattr(cls, name) else getattr(cls, name)
    
    
    @property
    def key(self):
        return getattr(self, '_key')
    
    @property
    def defaults(self) -> dict:
        defaults = {}
        for k,v in self.fields().items():
            if is_defined(v.factory):
                defaults.update({k: v.factory()})  
            elif is_defined(v.default):
                defaults.update({k: v.default})
        return defaults
    

class BaseField(ABC):
    POSITIONAL = POSITIONAL_HTML_ELEMENT_ATTRIBUTES
    KEYVALUE = KEYWORD_VALUE_HTML_ELEMENT_ATTRIBUTES
    CONFIG = [*BASE_FIELD_CONFIG_KEYS, *POSITIONAL_HTML_ELEMENT_ATTRIBUTES, *KEYWORD_VALUE_HTML_ELEMENT_ATTRIBUTES]
    
    def __init__(self, 
                 *args, 
                 required=True,
                 hidden=Variable.UNDEFINED,
                 disabled=Variable.UNDEFINED,
                 multiple=Variable.UNDEFINED,
                 readonly=Variable.UNDEFINED,
                 placeholder=Variable.UNDEFINED,
                 max=Variable.UNDEFINED,
                 min=Variable.UNDEFINED,
                 minlength=Variable.UNDEFINED,
                 maxlength=Variable.UNDEFINED,
                 step=Variable.UNDEFINED,
                 pattern=Variable.UNDEFINED,
                 transform=Variable.UNDEFINED,
                 ftype='text',
                 fsize=(12,6,3),
                 fclass=Variable.UNDEFINED,
                 fstyle=INPUT_STYLE,
                 tag='input',
                 default=Variable.UNDEFINED,
                 factory=Variable.UNDEFINED,
                 exclude=Variable.UNDEFINED,
                 title=Variable.UNDEFINED,
                 description=Variable.UNDEFINED,
                 choices=Variable.UNDEFINED,
                 **kwargs) -> None:
        self.required=required
        self.hidden=hidden
        self.disabled=disabled
        self.multiple=multiple
        self.readonly=readonly
        self.placeholder=placeholder
        self.max=max
        self.min=min
        self.minlength=minlength
        self.maxlength=maxlength
        self.step=step
        self.pattern=pattern
        self.ftype=ftype
        self.fclass=fclass
        self.fsize=fsize
        self.fstyle=fstyle
        self.transform=transform
        self.tag=tag
        self.default=default
        self.factory=factory
        self.title=title
        self.exclude=exclude
        self.description=description
        self.choices=choices
        self.element=None
        self._instance_ = None
        self._dbdata_ = None
        
        if len(args) > 0:
            if type(args[0]) == str:
                self.owner = None
                self.owner_name = args[0]
            else:
                self.owner = args[0]
                self.owner_name = self.owner.__name__
        else:
            self.owner = Variable.UNDEFINED
            self.owner_name = Variable.UNDEFINED
                
    def default_value(self):
        value = None
        if is_defined(self.factory):
            value = self.factory()
        elif is_defined(self.default):
            value = self.default
        if isinstance(value, BaseEnum):
            return value.name
        return value
    
    def submit_button(self):
        return str(Input(ftype='submit', style=SUBMIT_STYLE))

    def get_formfield_attrs(self, novalue=False) -> str:
        result = [f'id="{self.public_name}" name="{self.public_name}"']
        for k,v in vars(self).items():
            if is_defined(v):
                if k in self.KEYVALUE:
                    result.append(f'{k}="{v}"')
                elif k in self.POSITIONAL and v == True:
                    result.append(k)
        if is_defined(self.ftype):
            result.append(f'type="{self.ftype}"')
        if self.default_value():
            if not novalue:
                result.append(f'value="{str(self.default_value())}"')
        if is_defined(self.fstyle):
            result.append(f'style="{self.fstyle}"')
        if is_defined(self.fclass):
            result.append(f'class="{self.fclass} field"')
        else:
            result.append(f'class="field"')
        if is_defined(self.placeholder):
            result.append(f'placeholder="{self.placeholder}"')
        return ' '.join(result)
    
    async def get_formfield(self, inner_text=None):
        label = lambda field: f'<label for="{self.public_name}" class="form-label">{self.title if is_defined(self.title) else self.description if is_defined(self.description) else self.public_name} {field}</label>'
        fd = ''
        if self.element == Input:
            fd = f'<input {self.get_formfield_attrs()}>'
        elif self.element == DataList:
            lt = f"{self.public_name}_list"
            fd = f'<input list="{lt}" {self.get_formfield_attrs(novalue=True)}><datalist id="{lt}">{inner_text or await self.owner.options(selected=self.default_value())}</datalist>'
        elif self.element == Select:
            fd = f'<select {self.get_formfield_attrs(novalue=True)}>{inner_text or await self.owner.options(selected=self.default_value())}</select>'
        elif self.element == TextArea:
            fd = f'<textarea {self.get_formfield_attrs()}>{inner_text or ""}</textarea>'
        return label(fd)

    def pre_validate(self, obj, value) -> Any:
        if not exist(value):
            if is_defined(self.factory):
                value = self.factory()
            elif is_defined(self.default):
                value = self.default        
        return value
    
    def check_required(self, obj, value) -> None:
        if self.required == True:
            if not exist(value):
                raise ValueError(f'{type(obj).__name__}.{self.public_name} cannot be "{value}"')
            
    def post_validate(self, obj, value) -> Any:
        if exist(value):
            if is_defined(self.transform):
                if not issubclass(type(self), (EnumField, ReferenceField)):
                    return self.transform(value)
        return value
            
    def validate(self, obj, value) -> None:
        self.check_required(obj, value)
        if exist(value):
            if is_defined(self.min):
                if float(self.min) > value:    
                    raise ValueError(f'{self._name_} of {type(obj).__name__} is "{value}" and cannot be lesser then {self.min}')
            if is_defined(self.max):
                if float(self.max) < value:    
                    raise ValueError(f'{self._name_} of {type(obj).__name__} is "{value}" and cannot be greater then {self.max}')
            if is_defined(self.minlength):
                if float(self.minlength) > len(value):    
                    raise ValueError(f'{self._name_} of {type(obj).__name__} is "{value}" and cannot has length lesser then {self.minlength}')
            if is_defined(self.maxlength):
                if float(self.maxlength) < value:    
                    raise ValueError(f'{self._name_} of {type(obj).__name__} is "{value}" and cannot be greater then {self.maxlength}')
            if is_defined(self.pattern):
                if not re.match(self.pattern, value):
                    raise ValueError(f'{self._name_} of {type(obj).__name__} is "{value}" does not match {self.pattern}')
                             
    def __set__(self, obj, value) -> None:
        pre = self.pre_validate(obj, value)
        self.validate(obj, pre)
        post = self.post_validate(obj, pre)
        setattr(obj, self.private_name, post)
        
    def parse(self, obj, value) -> Any:
        return value 
    
    def unparsed(self, obj):
        '''Return attribute of object saved as private name. This is the value passed in to the __init__ method of the model.'''
        return getattr(obj, self.private_name)

    def __set_name__(self, obj, name) -> None:
        self.public_name = name 
        self.private_name = f'_{name}'
    
    def __get__(self, obj, owner=None) -> Any:
        return self.parse(obj, getattr(obj, self.private_name))
    

class  RegularField(BaseField):
    '''
    Used for direct access of database  
    '''
    pass 


class  EnumField(BaseField):
    '''
    Used for Enum classes lookup
    '''
    pass 

        
class ReferenceField(BaseField):
    '''
    Used to store the key and get access of database object
    '''
    pass 


class DateTime(RegularField):
    def __init__(self, **kwargs) -> None:
        super().__init__(datetime.datetime, **kwargs) 
        self.element = Input
        self.tag = 'input'
        self.ftype = 'datetime-local'
        
    def pre_validate(self, obj, value):
        value = super().pre_validate(obj, value)
        if isinstance(value, datetime.datetime):
            value = value.isoformat()
        return value
    
    def  parse(self, obj, value):
        if value:
            if isinstance(value, str):
                try:
                    value = datetime.datetime.fromisoformat(value)
                except BaseException as e:
                    print(e)
        return value 
        
class Text(RegularField):
    def __init__(self, **kwargs) -> None:
        super().__init__(str, **kwargs) 
        self.element = TextArea
        self.tag = 'textarea'
        self.ftype = Variable.UNDEFINED
        
    def post_validate(self, obj, value):
        value = super().post_validate(obj, value)
        return value 
    
    def parse(self, obj, value):
        if not isinstance(value, str):
            value = str(value)
        return value 

        
class SemicolonStrToList(RegularField):
    def __init__(self, **kwargs) -> None:
        super().__init__(str, **kwargs) 
        self.element = TextArea
        self.tag = 'textarea'
        self.ftype = Variable.UNDEFINED
        
    def post_validate(self, obj, value):
        value = super().post_validate(obj, value)
        if isinstance(value, (list, tuple)):
            value = '; '.join([str(x) for x in value])
        return value 
    
    def parse(self, obj, value):
        trim = lambda x: x.strip()
        split = lambda x: x.split(";")
        process = lambda x: [i for i in sorted([trim(w) for w in split(x) if w not in ['', None]]) if i not in ['', None]]
        if value:
            return process(value)
        return value

    
class String(RegularField):
    
    def __init__(self,**kwargs) -> None:
        super().__init__(str, **kwargs) 
        self.element = Input

    def parse(self, obj, value):
        return value
    
    def post_validate(self, obj, value):
        value = super().post_validate(obj, value)
        return value 
    
    def parse(self, obj, value):
        if not isinstance(value, str):
            value = str(value)
        return value 

        
class Choice(EnumField):
    '''
    Receive enumeration name to store in db. 
    The __get__ method return the enumeration member instance.
    '''
    
    def __init__(self,enum, **kwargs) -> None:
        super().__init__(enum, **kwargs) 
        self.element = Select
        self.tag = "select"
        self.ftype = Variable.UNDEFINED

        
    def  parse(self, obj, value):
        if value:
            return self.owner.get(value)
        return value 
    
class SelectEnum(EnumField):
    '''
    Receive enumeration name to store in db. 
    The __get__ method return the enumeration member instance.
    '''
    
    def __init__(self,enum, **kwargs) -> None:
        super().__init__(enum, **kwargs) 
        self.element = Select
        self.tag = "select"
        self.ftype = Variable.UNDEFINED

        
    def  parse(self, obj, value):
        if value:
            try:
                value = self.owner[value]
            except KeyError:
                try:
                    value = self.owner(value)
                except:
                    try:
                        value = self.owner.get(value)
                    except:
                        pass 
        return value 

    
class InputListEnum(EnumField):
    def __init__(self,enum, **kwargs) -> None:
        super().__init__(enum, **kwargs) 
        self.element = DataList
        self.tag = "datalist"
        self.ftype = Variable.UNDEFINED
        self.options = ''.join([str(Option(value=key, content=option)) for key, option in self.owner.choices()])

    def  parse(self, obj, value):
        if value:
            return self.owner.get(value)
        return value 

    

class Date(RegularField):
    '''
    Receive date as a isoformat string to store in db. 
    The __get__ method return the datetime.date instance.
    '''
    
    def __init__(self, **kwargs) -> None:
        super().__init__(datetime.date, **kwargs) 
        self.tag = 'input'
        self.ftype = 'date'
        self.element = Input


    def  parse(self, obj, value):
        return datetime.date.fromisoformat(value) if isinstance(value, str) else value  



class Title(RegularField):
    '''
    Receive a string to store in db.
    The __get__ method return the db string formated to title().
    '''
    
    def __init__(self, **kwargs) -> None:
        super().__init__(str, **kwargs) 
        self.element = Input

    def parse(self, obj, value):
        if isinstance(value, str):
            return value.title()
        return value 

    
    
class Number(RegularField):
    '''
    Receive a string or number store in db.
    The __get__ method return as a float.
    '''
    def __init__(self, **kwargs) -> None:
        super().__init__(float, **kwargs) 
        self.element = Input
        self.tag = 'input'
        self.ftype = 'number'

    def parse(self, obj, value):
        return float(value) if isinstance(value, str) else value  

    
    
class Integer(RegularField):
    '''
    Receive a string or number store in db.
    The __get__ method return as a integer.
    '''
    def __init__(self,**kwargs) -> None:
        super().__init__(int, **kwargs) 
        self.element = Input
        self.tag = 'input'
        self.ftype = 'number'
        
    def parse(self, obj, value):
        return int(value) if isinstance(value, str) else value  

    
    
class Float(RegularField):
    '''
    Receive a string or number store in db.
    The __get__ method return as a float.
    '''
    def __init__(self, **kwargs) -> None:
        super().__init__(float, **kwargs) 
        self.element = Input
        self.tag = 'input'
        self.ftype = 'number'
        
    def parse(self, obj, value):
        return float(value) if isinstance(value, str) else value  

    
    
class Real(RegularField):
    '''
    Receive a string or number store in db.
    The __get__ method return as a Decimal.
    '''
    def __init__(self,**kwargs) -> None:
        super().__init__(Decimal, **kwargs) 
        self.element = Input
        self.tag = 'input'
        self.ftype = 'number'
                
    def parse(self, obj, value):
        if value:
            return Decimal(value)
        return value

class Boolean(RegularField):
    '''
    Receive a int, string or bool and store as bool in db.
    '''
    def __init__(self,**kwargs) -> None:
        super().__init__(bool, **kwargs) 
        self.element = Input
        self.tag = 'input'
        self.ftype = 'checkbox'
        self.fstyle = 'display:inline-block'
        
    def pre_validate(self, obj, value):
        value = super().pre_validate(obj, value)
        return True if value in [1, 'true', True] else False
                
    def parse(self, obj, value):
        return True if value in [1, 'true', 'True', 'TRUE', True] else False

    

class Array(RegularField):
    '''
    Receive a string to stored in db.
    The __get__ method parse the string to a list.
    '''
    def __init__(self,**kwargs) -> None:
        super().__init__(list, **kwargs) 
        self.element = Input
        self.tag = 'input'
                
    def post_validate(self, obj, value):
        value = super().post_validate(obj, value)
        if isinstance(value, (list, tuple)):
            value = '; '.join([str(x) for x in value])
        return value 
    
    def parse(self, obj, value):
        if value:
            if isinstance(value, str):
                value = [x.strip() for x in re.split('[;\n]', value)]
        return value

    

class KeyData(ReferenceField):
    def __init__(self,table, project: Project = Project.CLIENT, fields=None, **kwargs) -> None:
        super().__init__(table, **kwargs) 
        self.project = project
        self.element = Select
        self.tag = 'select'
        self.ftype = Variable.UNDEFINED
        self.table = self.owner_name
        self.fields = fields or list()
        self.db = getattr(Connection(), str(project))
        
    def parse(self, obj, value):
        dbdata = self.db.sync_connect(self.table).get(str(value)) or dict()
        fields = self.fields if isinstance(self.fields, list) else [x.strip() for x in self.fields.split()]
        if fields:
            result = ChainMap({}, dbdata)
            result.update({"key": value, "table": self.table})
            for item in fields:
                result.update({item: result.get(item)})
            return result.maps[0]
        else:
            return dbdata
    
    
class DataField(ReferenceField):
    def __init__(self,table,  project: Project = Project.CLIENT, fields=None, **kwargs) -> None:
        super().__init__(table, **kwargs) 
        self.db = getattr(Connection(), str(project))
        self.table = self.owner_name
        self.element = DataList
        self.tag = 'datalist'
        self.ftype = Variable.UNDEFINED
        self.fields = fields or list()
        
    def parse(self, obj, value):
        dbdata = self.db.sync_connect(self.table).get(str(value)) or dict()
        fields = self.fields if isinstance(self.fields, list) else [x.strip() for x in self.fields.split()]
        if fields:
            result = ChainMap({}, dbdata)
            result.update({"key": value, "table": self.table})
            for item in fields:
                result.update({item: result.get(item)})
            return result.maps[0]
        else:
            return dbdata
        

class Instance(ReferenceField):
    '''Instance Field Class accept only the key of the object'''
    def __init__(self,owner, fields=None, **kwargs) -> None:
        super().__init__(owner, **kwargs)
        self.fields = fields if isinstance(fields, list) else [x.strip() for x in fields.split()] if isinstance(fields, str) else ''
#         if self.fields:
#             fields = 
#             self.ntuple = namedtuple(self.table, ' '.join(fields.keys()))
#         else:
#             self.ntuple = None
        self.project = self.owner.project()
        self.db = getattr(Connection(), str(self.project))
        self.element = Select
        self.tag = 'select'
        self.ftype = Variable.UNDEFINED
        self.table = self.owner_name
        self.fields = fields
        self._instance_ = None
        self._dbdata_ = None
        
    def parse(self, obj, value):
        def lazy_dbdata() -> dict:
            if self._instance_ and self._dbdata_:
                assert self._instance_.key == self._dbdata_['key']
            elif self._dbdata_ and value:
                assert value == self._dbdata_['key']
            elif exist(value):
                self._dbdata_ = self.db.sync_connect(self.table).get(str(value)) or dict()
        
        def lazy_instance():
            if not self._instance_:
                self._instance_ = self.owner(**self._dbdata_)
            return self._instance_
        
        lazy_dbdata()
        return lazy_instance() 
        

    

class Key(ReferenceField):
    def __init__(self,table,  project: Project = Project.CLIENT, **kwargs) -> None:
        super().__init__(table, **kwargs) 
        self.project = project
        self.element = Select
        self.tag = 'select'
        self.ftype = Variable.UNDEFINED
        self.table = self.owner_name
        
        
class ArrayOfKeys(ReferenceField):
    def __init__(self,table,  project: Project = Project.CLIENT, **kwargs) -> None:
        super().__init__(table, **kwargs) 
        self.project = project
        self.element = Select
        self.tag = 'select'
        self.multiple = True
        self.ftype = Variable.UNDEFINED
        self.table = self.owner_name
        
    def parse(self, obj, value):
        if isinstance(value, str):
            value = [x.strip() for x in re.split('\s', value)]
        elif isinstance(value, list):
            pass 
        return value
    

class BaseManager(ABC):
    
    def __init__(self, user: User = None, connection: Connection = None) -> None:
        self.user = user
        self.connection = connection or Connection()
    
    
    async def get_form(self, model: BaseModel, method='get', action=".", extra_models=None):
        result= []
        for field in model.fields().values():
            result.append(await field.get_formfield())
            result.append('<br>')
        if extra_models:
            for extra_model in extra_models:
                extra_result = []
                for extra_field in extra_model.fields().values():
                    extra_result.append(await extra_field.get_formfield())
                    extra_result.append('<br>')
                result.append(str(Fieldset(legend=extra_model.singular(), content=''.join(extra_result))))
        return str(Form(pk=f'{model.__name__}Form', content=str(Fieldset(legend=model.singular(), content=result)), method=method, action=action, novalidate=True)) 
    
    @cache
    async def model_instances(self, model):
        async with self.db(Project(model.project())).ListAll(model.table()) as result:
            return [model(**item) for item in result]
    
    @cache
    async def namedtuple_instances(self, model):
        ntuple = model.namedtuple()
        async with self.db(Project(model.project())).ListAll(model.table()) as result:
            return [ntuple(**item) for item in result]
        
    @property
    def clientdb(self):
        return self.connection.client
        
    @property
    def logdb(self):
        return self.connection.log
    
    @property
    def basedb(self):
        return self.connection.base
    
    @property
    def userdb(self):
        return self.connection.user
    
    def db(self, project: Project = Project.BASE):
        return {
            'base': self.basedb,
            'client': self.clientdb,
            'log': self.logdb,
            'user': self.userdb,
        }.get(str(project), self.basedb)

    @staticmethod 
    def table(instance: BaseModel):
        return instance.table() or type(instance).__name__
    
    @staticmethod 
    def json(instance: BaseModel):
        return instance.json()
    
    @staticmethod 
    def project(instance: BaseModel):
        return instance.project() or 'base'
    
    @cache
    async def enum_from_db(self, model: BaseModel):
        return BaseEnum(f'{model}Enum', [(x.key, x) for x in await model.list_all()])
    
    @cache
    def sync_enum_from_db(self, model: BaseModel):
        return asyncio.run(self.enum_from_db(model))

    
    
class Manager(BaseManager):
    '''Subclass of BaseManager'''


from metadoctor.base import *
from metadoctor.enumeration import *
from metadoctor import utils 

age = utils.age
now = utils.now 
today = utils.today
only_digits = utils.only_digits

class Person(BaseModel):
    """
    Profile.Person Class
    """
    __project = Project.USER
    __singular = 'Pessoa'
    __plural = 'Pessoas'

    fullname = Title(title='Nome completo')

    class Gender(BaseEnum):
        M = 'Masculino'
        F = 'Feminino'
        N = 'Não Binário'

    gender = Choice(Gender, title='Região')
    birthdate = Date(title='Nascimento')
    cpf = String(title='CPF', required=False)

    def __init__(self, fullname, gender, birthdate, cpf=None, **kwargs) -> None:
        super().__init__(fullname=fullname, gender=gender, birthdate=birthdate, cpf=cpf, **kwargs)
        if self.cpf:
            self.cpf = only_digits(self.cpf)
            assert len(self.cpf) == 11, 'CPF deve ter exatamente 11 dígitos. ' + str(len(self.cpf)) + ' foram encontrados.'

    @property
    def name(self):
        return self.fullname 

    @property
    def bdate(self):
        return self.birthdate 
    
    
    @property
    def age(self):
        return age(self.birthdate)
    
    
class User(BaseModel):
    """
    User Class
    """
    __project = Project.USER
    __singular = 'Usuário'
    __plural = 'Usuários'

    username: str = String(default='Indefinido', title='Nome de Usuário')
    is_admin = Boolean(title='Perfil Administrador', readonly=True)
    created = DateTime(factory=lambda: now().isoformat(), exclude=True, readonly=True)

    def __init__(self, username, is_admin=False, created=None, **kwargs) -> None:
        super().__init__(username=username, is_admin=is_admin, created=created,  **kwargs)
        
        
    def _autokey(self):
        return remove_accents(self.username)
        
class Profile(BaseModel):
    """
    Profile Class
    """
    __project = Project.USER
    __singular = 'Perfil'
    __plural = 'Perfis'
    
    class ProfileType(BaseEnum):
        DOCTOR = 'Médico'
        THERAPIST = 'Terapeuta'
        EMPLOYEE = 'Funcionário'
        ASSISTANT = 'Assistente'
        PATIENT = 'Paciente'
        RELATIVE = 'Familiar'

    type =  SelectEnum(ProfileType, default=ProfileType.PATIENT, title='Tipo de Perfil')
    person = Instance(Person, title='Pessoa')
    user = Instance(User, required=False, title='Usuário')
    
    def __str__(self):
        return f'{self.person.fullname} ({self.type})'
    
class PatientInfo(BaseModel):
    '''Profile.Patient Class'''
    __project = Project.CLIENT
    __pt_singular = 'Paciente'
    __plural = 'Pacientes'

    relatives = ArrayOfKeys('Profile', 'user', title='Familiares', required=False)
    health_plan = SemicolonStrToList(title='Planos de Saúde', required=False)
    curator = Key('Profile', 'user', required=False)
    notes = Text(required=False)

    def __init__(self, relatives=None, health_plan=None, curator=None,notes=None, **kwargs):
        super().__init__(relatives=relatives, health_plan=health_plan, curator=curator, notes=notes, **kwargs)
        

    
    
class ProviderInfo(BaseModel):
    '''Profile.Provider Class'''
    __project = Project.USER
    __singular = 'Profissional'
    __plural = 'Profissionais'

    class Graduation(BaseEnum):
        MED = 'Medicina'
        PSY = 'Psicologia'
        FIS = 'Fisioterapia'
        NUR = 'Enfermagem'
        FON = 'Fonoaudiologia'
        
    profile = Instance(Profile)

    graduation = Choice(Graduation, default=Graduation.MED)
    licence = String(title='Registro Profissional')
    specialties = Array(required=False, exclude=True)
    notes = Text(required=False, exclude=True)

    def __init__(self, graduation, licence, specialties=None, notes=None, **kwargs):
        super().__init__(graduation=graduation, licence=licence, specialties=specialties, notes=notes, **kwargs)
        
    
class EmployeeInfo(BaseModel):
    '''Profile.Employee Class'''
    __project = Project.CLIENT
    __singular = 'Funcionário'
    __plural = 'Funcionários'
    
    profile = Instance(Profile)
    functions = SemicolonStrToList()
    salary = Real(title='Salário', exclude=True)
    skills = SemicolonStrToList(required=False)
    notes = Text(required=False)
            
        
class Address(BaseModel):
    """
    Profile.Address Class
    """
    __project = Project.CLIENT
    __singular = 'Endereço'
    __plural = 'Endereços'

    class City(BaseModel):
        __project = Project.BASE
        __singular = 'Cidade'
        __plural = 'Cidades'
        __table = 'City'

        name = Title()
        region = KeyData('Region', 'base')
        
    profile = Instance(Profile)
    city = Instance(City, title='Cidade')
    cep = String(title='CEP', required=False)
    street = Title(title='Logradouro', required=False)
    number = String(title='Número', required=False)
    complement = String(title='Complemento', required=False)
    district = Title(title='Bairro', required=False)
    notes = Text(required=False)

    def __init__(self, profile, city, cep=None, street=None, number=None, complement=None, district=None,  notes=None, **kwargs) -> None:
        super().__init__(profile=profile, city=city, cep=cep, street=street, number=number, complement=complement, district=district,  notes=notes, **kwargs)
        
class Contact(BaseModel):
    """
    Profile.Contact Class
    """
    __project = Project.CLIENT
    __singular = 'Contato'
    __plural = 'Contatos'
    
    class Type(StrEnum):
        TELEPHONE = 'Telefone Fixo'
        MOBILE = 'Telefone Celular'
        EMAIL = 'Email Pessoal'
    
    profile = Instance(Profile)
    type = Choice(Type, default=Type.MOBILE)
    value = String()
    notes = Text(required=False)
    
    
    def __init__(self, profile, type, value, notes=None, **kwargs):
        super().__init__(profile=profile, type=type, value=value, notes=notes, **kwargs)

            
class Log(BaseModel):
    '''Log Register class'''
    __project = Project.LOG
    __singular = 'Registro'
    __plural = 'Registros'
    
    datetime = DateTime(factory=now)
    profile = Instance(Profile)
    path = String()
    method = SelectEnum(Profile.ProfileType)
    result = String()


__all__ = [
    'Settings',
]

import os
from collections import ChainMap
from starlette.config import Config
from starlette.datastructures import Secret
from dataclasses import dataclass



WORKING_DIRECTORY = os.getcwd()
ENV_FILE = os.path.abspath('.env')

try:
    config = Config(env_file=ENV_FILE)
except:
    config = Config()



@dataclass
class Settings:
    config: Config = config
        
    def __post_init__(self):
        self.chain = ChainMap({}, {k: Secret(v) for k,v in vars(self.config)['file_values'].items()})
        
    @property
    def cwd(self):
        return os.getcwd()
    
    @property
    def env_file(self):
        return os.path.abspath('.env')

    def update(self, data: dict):
        self.chain.update(data)

    def get(self, key):
        return self.chain.get(key)
    
    def all_chain(self):
        return self.chain
    
    def initial(self):
        return {k:Secret(v) for k,v in self.chain.maps[-1].items()}    
    


#!/usr/bin/env python
# coding: utf-8

"""BaseEnum module"""

__all__ = [
    'Variable',
    'BaseEnum',
    'StrEnum', 
    'TupleEnum', 
    'Day', 
    'Frequency', 
    'Month', 
    'Permission',
    'Right',
    'Access',
    'Gender',
    'VisitType',
    'ActiveDrugType',
    'Pharmaceutical',
    'RecipeType',
    'DosageUnit',
    'DosageForm',
    'DosageUse',
    'ActiveDrug',
    'Brand',
    'Drug',
    'Experience',
    'BodySystem',
    'Profession',
    'PaymentMethod',
    'Percent',
    'Cowntry',
    'Project',
    ]

from enum import Enum 
from datetime import timedelta, date, datetime
from typing import Union
from collections import ChainMap
from functools import cache

from metadoctor import utils


class Variable(Enum):
    UNDEFINED = 'Indefinido'
    MISSING = 'Ausente'
    
    def __str__(self):
        return self.value
    
    def __repr__(self):
        return self.value

    @staticmethod
    def is_defined(value):
        return False if value == Variable.UNDEFINED else True
    
    @staticmethod
    def is_missing(value):
        return True if value == Variable.MISSING else False
    
    @staticmethod
    def exist(value):
        return False if value in ['', None, Variable.MISSING, Variable.UNDEFINED] else True

class BaseEnum(Enum):
    
    def __str__(self) -> str:
        if type(self.value) == str:
            return self.value
        elif type(self.value) == int:
            return str(self.value)
        elif isinstance(self.value, (tuple, list)):
            gen = (x for x in self.value if isinstance(x, str))
            return next(gen)
        else:
            return str(self.value)
    
    def __repr__(self) -> str:
        return f'{type(self).__name__}.{self.name}'
    
    
    def __html_option__(self):
        return f'<option value="{self.name}">{str(self)}</option>'
    
    
    @classmethod
    def members(cls):
        return [e for e in cls.__members__.values()]
    
    @classmethod
    def choices(cls, selected=None):
        members = cls.members()
        choices = []
        for item in members:
            if item == selected or item.name == selected or item.value == selected or str(item) == str(selected):
                choices.append((item.name, item, True))
            else:
                choices.append((item.name, item, False))
        return choices
    
    @classmethod
    async def options(cls, selected=None):
        class Option:
            def __init__(self, value, content, selected=False):
                self.value = value
                self.content = content 
                self.selected = selected 
                
            def __str__(self):
                return f'<option value="{self.value}">{self.content}</option>' if self.selected == False else f'<option value="{self.value}" selected>{self.content}</option>'
            
        members = cls.members()
        options = []
        for item in members:
            if item == selected or item.name == selected or item.value == selected:
                options.append(Option(value=item.name, content=item.value, selected=True))
            else:
                options.append(Option(value=item.name, content=item.value))
        return ''.join([str(x) for x in options])
    

    @classmethod
    def name2member_map(cls):
        '''member name as keyword and member as value'''
        return {**{e.name: e for e in cls.members()}, ** {utils.remove_accents(e.name).lower(): e for e in cls.members()}}
    
    @classmethod
    def str2member_map(cls):
        '''normalized and lower str of member as keyword and member as value'''
        return {**{str(e): e for e in cls.members()}, **{utils.remove_accents(str(e)).lower(): e for e in cls.members()}}
    
    
    @classmethod
    def int2member_map(cls):
        '''int as keyword and member as value'''
        try:
            return {int(e): e for e in cls.members()}
        except:
            return {}
        
    @classmethod
    def tuple2member_map(cls):
        '''int as keyword and member as value'''
        try:
            return {tuple(e): e for e in cls.members()}
        except:
            return {}
    
    @classmethod
    def value2member_map(cls):
        '''member value as keyword and member as value'''
        return {**{e.value: e for e in cls.members()}, ** {utils.remove_accents(str(e.value)).lower(): e for e in cls.members()}}
    
    
    @classmethod
    def chain_map(cls):
        return ChainMap(
            cls.name2member_map(), 
            cls.str2member_map(), 
            cls.tuple2member_map(), 
            cls.value2member_map(), 
            cls.int2member_map()
        )
    
    @classmethod
    def get(cls, v: Union[str, int]):
        return cls.chain_map().get(utils.remove_accents(v).lower() if isinstance(v, str) else v)


class StrEnum(str, BaseEnum):
    '''
    StrEnum
    '''

    
class TupleEnum(tuple, BaseEnum):
    '''
    TupleEnum
    `````````
    '''

    def __str__(self) -> str:
        return str(self.value[1])
    
    def __int__(self) -> int:
        return self.value[0]


class Day(timedelta, BaseEnum):
    '''
    Day enumerator 
    '''
    _ignore_ = 'Day item'
    Day = vars()
    for item in range(366):
        Day[f'{item}'] = item
        
    def __str__(self) -> str:
        days = self.value.days
        return f'{days} {"dia" if days <= 1 else "dias"}'
    
    @classmethod
    def __date_ahead__(cls, v: int):
        return utils.today() + cls[str(v)].value
    
    @classmethod
    def __date_ago__(cls, v: int):
        return utils.today() - cls[str(v)].value
    
    @classmethod
    def value2member_map(cls):
        members = cls.__members__.values()
        return {**{e.value.days: e for e in members}, **{e.value: e for e in members}}


class Month(TupleEnum):
    JAN = 1, "January", "Janeiro", "Jan"
    FEB = 2, "February", "Fevereiro", "Fev"
    MAR = 3, "March", "Março", "Mar"
    APR = 4, "April", "Abril", "Abr"
    MAY = 5, "May", "Maio", "Mai"
    JUN = 6, "June", "Junho", "Jun"
    JUL = 7, "July", "Julho", "Jul"
    AUG = 8, "August", "Agosto", "Ago"
    SEP = 9, "September", "Setembro", "Set"
    OCT = 10, "October", "Outubro", "Out"
    NOV = 11, "November", "Novembro", "Nov"
    DEZ = 12, "December", "Dezembro", "Dez"
    
    def __int__(self):
        return self.value[0]

    def __str__(self):
        return self.value[2]
    
    def __tuple__(self):
        return self.value


class Frequency(TupleEnum):
    '''
        Frequency Enum 
            tuple type enumeration
            properties: title display hours 
    '''
    NO = 'Nenhuma', 'nenhuma', 0, False
    QAM = 'Manhã', 'pela manhã', 24, True
    QPM = 'Tarde', 'à tarde', 24, True
    QHS = 'Noite', 'à noite', 24, True
    Q1H = '1/1h', 'a cada 1 horas', 1, True
    Q2H = '2/2h', 'a cada 2 horas', 2, True
    Q3H = '3/3h', 'a cada 3 horas', 3, True
    Q4H = '4/4h', 'a cada 4 horas', 4, True
    Q6H = '6/6h', 'a cada 6 horas', 6, True
    Q8H = '8/8h', 'a cada 8 horas', 8, True
    Q12H = '12/12h', 'a cada 12 horas', 12, True
    QD = '24/24h', 'a cada 24 horas', 24, True
    QOD = 'Dias Alternados', 'a cada 2 dias', 48, True
    QW = 'Semanal', 'por semana', 24*7, True
    QOW = 'Duas Semanas', 'a cada duas semana', 2*24*7, True
    QM = 'Mensal', 'ao mês', 24*30, True
    QOM = 'Bimestral', 'a cada dois meses', 2*24*30, True
    QY = 'Anual', 'ao ano', 24*365, True
    UD = 'Como Instruído', 'como instruído', 0, True 
    PRN = 'Caso Necessário', 'caso necessário', 0, True 

    def __str__(self) -> str:
        return self.value[0]
    
    def __int__(self):
        return self.value[2]

    def __bool__(self):
        return self.value[3]
    
    def __tuple__(self):
        return self.value
    
    @classmethod
    def value2member_map(cls):
        members = cls.__members__.values()
        return {**{e.value[0]: e for e in members}, **{e.value[1]: e for e in members}, **{(e.value[2],e.value[3]): e for e in members}}


class Permission(StrEnum):
    A = 'Permitido'
    F = 'Proibido'
    U = 'Indefinido'
    
    
class Right(StrEnum):
    R = 'Ler'
    W = 'Escrever'
    
    
class Access(StrEnum):
    P = 'Público'
    U = 'Usuário'
    A = 'Administrador'   
    
    
class Gender(StrEnum):
    M = 'Masculino'
    F = 'Feminino'
    O = 'Outro'
    N = 'Nenhum'
    

class VisitType(StrEnum):
    I = 'Inicial'
    F = 'Seguimento' # follow-up
    R = 'Retorno' # revisit 
    S = 'Sessão Individual'
    G = 'Sessão em Grupo'
    

class ActiveDrugType(TupleEnum):
    '''ActiveDrugType: name'''
    BENZODIAZEPINE = 'Benzodiazepínico'
    ANTIPSYCHOTIC = 'Antipsicótico'
    ANTIDEPRESSANT = 'Antidperessivo'
    PSYCHOSTIMULANT = 'Psicoestimulante'
    MOOD_STATILIZER = 'Estabilizador do Humor'
    NON_BENZO_ANXIOLYTIC = 'Ansiolítico Não Benzodiazepínico'
    NON_BENZO_HYPNOTIC = 'Hipnótico'
    ANTI_HYPERTENSIVE = 'Antihipertensiva'
    DIURETIC = 'Diurético'
    ANTICONVULSIVE = 'Anticonvulsivante'
    VASODILATOR = 'Vasodilatador'


class Pharmaceutical(StrEnum):
    '''Pharmaceutical: name'''
    BIOLAB = 'Biolab'
    PFIZER = 'Pfizer'
    TAKEDA = 'Takeda'
    LUNDBECK = 'Lundbeck'
    ABBOT = 'Abbot'
    NOVARTIS = 'Novartis'
    LIBBS = 'Libbs'
    EUROFARMA = 'Eurofarma'
    ACHE = 'Aché'
    CRISTALIA = 'Cristália'
    EMS = 'EMS'
    SUPERA = 'Supera'


class RecipeType(TupleEnum):
    '''RecipeType: singular, plural, abbrev'''
    A = 'Receita A', 'Receitas A', 'A'
    B1 = 'Receita B1', 'Receitas B1', 'B1'
    B2 = 'Receita B2', 'Receitas B2', 'B2'
    C = 'Receita C', 'Receitas C', 'C'
    S = 'Receita Simples', 'Receitas Simples', 'S'


class DosageUnit(TupleEnum):
    '''DosageUnit: singular, plural, abbrev'''
    MG = 'miligrama', 'miligramas', 'mg'
    ML = 'mililitro', 'mililitros', 'mL'
    DL = 'decilitro', 'decilitros', 'dL'
    MG_PER_ML = 'miligrama por mililitro', 'miligramas por mililitro', 'mg/mL'
    MG_PER_DL = 'miligrama por decilitro', 'miligramas por decilitro', 'mg/dL'
    MCG = 'micrograma', 'microgramas', 'mcg'
    
    def __str__(self):
        return self.value[2]
    
    @property
    def abbrev(self):
        return self.value[2]


class DosageForm(TupleEnum):
    '''DosageForm: singular plural abbrev abbrev_plural'''
    CAPSULE = 'cápsula', 'cápsulas', 'cap', 'caps'
    TABLET = 'comprimido', 'comprimidos', 'cp', 'cps'
    SUBLINGUAL_TABLET = 'comprimido sublingual', 'comprimidos sublinguais', 'cp', 'cps'
    ORODISPERSIBLE_TABLET = 'comprimido orodispersível', 'comprimidos orodispersíveis', 'cp', 'cps'
    MILILITER = 'mililitro', 'mililitros', 'ml', 'ml'
    
    def __str__(self):
        return self.abbrev

    @property
    def abbrev(self):
        return self.value[2]


class DosageUse(TupleEnum):
    '''DosageUse: singular plural abbrev abbrev_plural conversion'''
    DOSAGE_FORM = None, None, None, None, 1
    DROP = 'gota', 'gotas', 'gt', 'gts', 20
    MICRODROP = 'microgota', 'microgotas', 'mcgt', 'mcgts', 60


class ActiveDrug(TupleEnum):
    '''ActiveDrug: name type recipe'''
    ALPRAZOLAM = 'Alprazolam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    BROMAZEPAM = 'Bromazepam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    CHLORDIAZEPOXIDE = 'Clordiazepóxido', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    CLONAZEPAM = 'Clonazepam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    DIAZEPAM = 'Diazepam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    FLURAZEPAM = 'Flurazepam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    FLUNITRAZEPAM = 'Flunitrazepam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    LORAZEPAM = 'Lorazepam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    MIDAZOLAM = 'Midazolam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    ESTAZOLAM = 'Estazolam', ActiveDrugType.BENZODIAZEPINE, RecipeType.B1
    VALPROIC_ACID = 'Ácido Valpróico', ActiveDrugType.MOOD_STATILIZER, RecipeType.C
    LAMOTRIGINE = 'Lamotrigina', ActiveDrugType.MOOD_STATILIZER, RecipeType.C
    CARBAMAZEPINE = 'Carbamazepina', ActiveDrugType.MOOD_STATILIZER, RecipeType.C
    DIVALPROEX_SODIUM = 'Divalproato de Sódio', ActiveDrugType.MOOD_STATILIZER, RecipeType.C
    LITHIUM_CARBONATE = 'Carbonato de Lítio', ActiveDrugType.MOOD_STATILIZER, RecipeType.C
    ARIPIPRAZOLE = 'Aripiprazol', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    QUETIAPINE = 'Quetiapina', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    LURASIDONE = 'Lurasidona', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    ZIPRASIDONE = 'Ziprasidona', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    RISPERIDONE = 'Risperidona', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    HALOPERIDOL = 'Haloperidol', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    CLOZAPINE = 'Clozapina', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    OLANZAPINE ='Olanzapina', ActiveDrugType.ANTIPSYCHOTIC, RecipeType.C
    VORTIOXETINE = 'Vortioxetina', ActiveDrugType.ANTIDEPRESSANT, RecipeType.C
    LISDEXAMFETAMINE = 'Lisdexanfetamina', ActiveDrugType.PSYCHOSTIMULANT, RecipeType.A
    METHYLPHENIDATE = 'Metilfenidato', ActiveDrugType.PSYCHOSTIMULANT, RecipeType.A
    BUSPIRONE = 'Buspirona', ActiveDrugType.NON_BENZO_ANXIOLYTIC, RecipeType.C
    ZOLPIDEM = 'Zolpidem', ActiveDrugType.NON_BENZO_HYPNOTIC, RecipeType.C
    DULOXETINE = 'Duloxetina', ActiveDrugType.ANTIDEPRESSANT, RecipeType.C
    LOSARTAN = 'Losartana', ActiveDrugType.ANTIDEPRESSANT, RecipeType.S
    HYDROCHLOROTHIAZIDE = 'Hidroclorotiazida', ActiveDrugType.DIURETIC, RecipeType.S
    BUPROPION = 'Bupropiona', ActiveDrugType.ANTIDEPRESSANT, RecipeType.C
    MIRTAZAPINE = 'Mirtazapina', ActiveDrugType.ANTIDEPRESSANT, RecipeType.C
    PREGABALIN = 'Pregabalina', ActiveDrugType.NON_BENZO_ANXIOLYTIC, RecipeType.C
    DESVENLAFAXINE = 'Desvenlafaxina', ActiveDrugType.ANTIDEPRESSANT, RecipeType.C
    TADALAFIL = 'Tadalafila', ActiveDrugType.VASODILATOR, RecipeType.S
    ESCITALOPRAM = 'Escitalopram', ActiveDrugType.ANTIDEPRESSANT, RecipeType.C


class Brand(TupleEnum):
    '''Brand: name drugs presentations dosage_unit dosage_form dosage_usage pharmaceutical'''
    BRINTELLIX = 'Brintellix', [ActiveDrug.VORTIOXETINE], [(5,30),(10,30),(10,60),(15,30),(15,60),(20,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.LUNDBECK
    VENVANSE = 'Venvanse', [ActiveDrug.LISDEXAMFETAMINE], [(30,28),(50,28),(70,28)], DosageUnit.MG, DosageForm.CAPSULE, DosageUse.DOSAGE_FORM, Pharmaceutical.TAKEDA
    RITALIN = 'Ritalina', [ActiveDrug.METHYLPHENIDATE], [(10,20),(10,30),(10,60)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.NOVARTIS
    RITALINALA = 'Ritalina LA', [ActiveDrug.METHYLPHENIDATE], [(10,30),(20,30),(30,30),(40,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.NOVARTIS
    ARPEJO = 'Arpejo', [ActiveDrug.ARIPIPRAZOLE], [(20,30)], DosageUnit.MG_PER_ML, DosageForm.MILILITER, DosageUse.DROP, Pharmaceutical.EMS
    CYMBI = 'Cymbi', [ActiveDrug.DULOXETINE], [(30,30),(30,60),(60,30),(60,60)], DosageUnit.MG, DosageForm.CAPSULE, DosageUse.DOSAGE_FORM, Pharmaceutical.EMS
    VELIJA = 'Velija', [ActiveDrug.DULOXETINE], [(30,30),(30,60),(60,30),(60,60)], DosageUnit.MG, DosageForm.CAPSULE, DosageUse.DOSAGE_FORM, Pharmaceutical.LIBBS
    BUPIUMXL = 'Bupium XL', [ActiveDrug.BUPROPION], [(150,30),(150,60),(300,30),(300,60)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EMS
    BUPXL = 'Bup XL', [ActiveDrug.BUPROPION], [(150,30),(300,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    ZETRONXL = 'Zetron XL', [ActiveDrug.BUPROPION], [(150,30),(150,60),(300,30),(300,60)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM,Pharmaceutical.LIBBS
    JUNEVE = 'Juneve', [ActiveDrug.LISDEXAMFETAMINE], [(30,28),(50,28),(70,28)], DosageUnit.MG, DosageForm.CAPSULE, DosageUse.DOSAGE_FORM, Pharmaceutical.TAKEDA
    MENELAT = 'Menelat', [ActiveDrug.MIRTAZAPINE], [(30,28),(50,28),(70,28)], DosageUnit.MG, DosageForm.CAPSULE, DosageUse.DOSAGE_FORM, Pharmaceutical.TAKEDA
    APICE = 'Ápice', [ActiveDrug.PREGABALIN], [(75,30),(150,30)], DosageUnit.MG, DosageForm.CAPSULE, DosageUse.DOSAGE_FORM, Pharmaceutical.SUPERA
    ARADOISH = 'Aradois H', [ActiveDrug.LOSARTAN, ActiveDrug.HYDROCHLOROTHIAZIDE], [((50,12.5),30),((50,12.5),60)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.BIOLAB
    PATZSL = 'Patz Sl', [ActiveDrug.ZOLPIDEM], [(5,20), (5,30), (5,60)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.BIOLAB
    PRISTIQ = 'Pristiq', [ActiveDrug.DESVENLAFAXINE], [(5,20), (5,30), (5,60)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.BIOLAB
    TADALAFIL = None, [ActiveDrug.TADALAFIL], [(5,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    ATTENZE = 'Attenze', [ActiveDrug.METHYLPHENIDATE], [(10,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    CARBOLITIUM = 'Carbolitium', [ActiveDrug.LITHIUM_CARBONATE], [(300,50)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    CARBOLITIUMCR = 'Carbolitium CR', [ActiveDrug.LITHIUM_CARBONATE], [(450,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    ESC = 'ESC', [ActiveDrug.ESCITALOPRAM], [(20,60),(10,60),(15,60),(20,30),(10,30),(15,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    ESCSOL = 'ESC Gotas', [ActiveDrug.ESCITALOPRAM], [(15,20)], DosageUnit.MG_PER_ML, DosageForm.MILILITER, DosageUse.DROP, Pharmaceutical.EUROFARMA
    QUETXR = 'Quet XR', [ActiveDrug.QUETIAPINE], [(50,30),(200,30),(300,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    QUET = 'Quet', [ActiveDrug.QUETIAPINE], [(25,30),(200,30),(100,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA
    ATIP = 'Atip', [ActiveDrug.QUETIAPINE], [(25,30),(200,30),(100,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.SUPERA
    ATIPXR = 'Atip XR', [ActiveDrug.QUETIAPINE], [(50,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.SUPERA
    ARISTAB = 'Aristab', [ActiveDrug.ARIPIPRAZOLE], [(10,30),(15,30),(20,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.ACHE
    TURNOSL = 'Turno Sl', [ActiveDrug.ZOLPIDEM], [(5,30)], DosageUnit.MG, DosageForm.TABLET, DosageUse.DOSAGE_FORM, Pharmaceutical.EUROFARMA

    @property
    def brand_name(self):
        val = self.value[0]
        return val if not val in ['', None] else 'Genérico'
    
    @property
    def drugs(self):
        return '/'.join([x.value[0] for x in self.value[1]]) 
    
    @property
    def presentations(self):
        result = []
        for item in self.value[2]:
            if isinstance(item[0], tuple):
                total = '+'.join([f"{x}{self.dosage_unit}" for x in item[0]])
            else:
                total = f'{item[0]}{self.dosage_unit}'
            result.append((self.brand_name, self.drugs, total, f"{item[1]}{self.dosage_form}", self.dosage_usage))
        return result 
    
    @property
    def dosage_unit(self):
        return self.value[3].abbrev
    
    @property
    def dosage_form(self):
        return self.value[4].abbrev
    
    @property
    def dosage_usage(self):
        return self.value[5].value[0] or self.dosage_form
    
    @property
    def pharmaceutical(self):
        return self.value[6].value[0]
    

class Drug(TupleEnum):
    _ignore_ = 'Drug brand Brand item name'
    Drug = vars()
    for brand in Brand.__members__.values():
        for item in brand.presentations:
            name = utils.remove_accents(f'{item[0]}{item[2]}{item[3]}').upper().replace(" ","").replace("/","").replace(".", "&").replace("+","")
            Drug[name] = item[0], item[1], item[2], item[3]
            
    def __str__(self):
        if self.value[0] in ['Genérico']:
            return f"{self.value[1]} {self.value[2]} {self.value[3]}"
        return f"{self.value[0]} ({self.value[1]}) {self.value[2]} {self.value[3]}"


class Experience(TupleEnum):
    _ignore_ = 'Experience data i start index names'
    Experience = vars()
    names = ['N3', 'N2', 'N1', 'E0', 'P1', 'P2', 'P3']
    data = ['Péssima', 'Muito Negativa', 'Negativa', 'Indiferente', 'Positiva', 'Muito Positiva', 'Excelente']
    start = -3
    index = 0
    Experience['UN'] = None, '---'
    for i in names:
        Experience[i] = start, data[index]
        start += 1
        index += 1
        
Experience.str = lambda x: f"{x.value[1]}"
Experience.int = lambda x: f"{x.value[0]}"


class BodySystem(StrEnum):
    CAR = 'Cardiovascular'
    RES = 'Respiratório'
    NEU = 'Nervoso'
    CIR = 'Circulatório'
    MSK = 'Musculoesquelético'
    DIG = 'Digestivo'
    SEX = 'Reprodutor'
    AUD = 'Auditório'
    VIS = 'Visual'
    TEG = 'Tegumentar'
    END = 'Endocrinológico'
    EXC = 'Excretor'
    
    
class PaymentMethod(StrEnum):
    CC = 'Cartão de Crédito'
    CD = 'Cartão de Débito'
    TR = 'Transferência Bancária'
    DI = 'Dinheiro'
    CH = 'Cheque'
    PX = 'Pix'
    
    
class Profession(StrEnum):
    MED = 'Medicina'
    PSY = 'Psicologia'
    FIS = 'Fisioterapia'
    NUR = 'Enfermagem'
    FON = 'Fonoaudiologia'
    OCP = 'Terapia Ocupacional'
    
    
class Percent(BaseEnum):
    _ignore_ = 'Percent val'
    Percent = vars()
    for val in range(101):
        Percent[f'{val}'] = f'{val}%', val/100
        
    def __float__(self):
        return self.value[1]
    
    def __str__(self):
        return self.value[0]
    
    
class Cowntry(BaseEnum):
    BRA = 'Brasil'
    USA = 'Estados Unidos'
    QAT = 'Qatar'
    POR = 'Portugal'
    ARG = 'Argentina'
    
    
class Project(StrEnum):
    BASE = 'base'
    LOG = 'log'
    CLIENT = 'client'
    USER = 'user'



#!/usr/bin/env python
# coding: utf-8
'''db module'''

__all__ = [
    'AbstractDB', 
    'Database', 
    'Connection',
]

import contextlib
from abc import ABC
from deta import Deta
from typing import Dict, Any
from starlette.datastructures import Secret

from metadoctor import utils
from metadoctor import settings
from metadoctor import enumeration

Project = enumeration.Project


class AbstractDB(ABC):
    
    def __init__(self, key, project=Project.BASE):
        self.key = key
        self.project = project

    def sync_connect(self, table: str):
        return Deta(str(self.key)).Base(table)

    async def connect(self, table: str):
        return Deta(str(self.key)).Base(table)

    @contextlib.asynccontextmanager
    async def Count(self, table):
        db = await self.connect(table)
        count = 0
        try:
            count = len(db.fetch().items)
        finally:
            yield count
            db.client.close()

    @contextlib.asynccontextmanager
    async def ListAll(self, table):
        db, result = await self.connect(table), []
        try:
            result = db.fetch().items
        except BaseException as e:
            print(e.__dict__)
        finally:
            yield result
            db.client.close()

    @contextlib.asynccontextmanager
    async def Get(self, table, key=None, keys=None):
        db, result = await self.connect(table), None
        try:
            if key:
                result = db.get(key=key)
            elif keys:
                result = []
                for key in keys:
                    result.append(db.get(key=key))
        finally:
            yield result
            db.client.close()

    @contextlib.asynccontextmanager
    async def Search(self, table, query=None):
        query = query or {}
        db = await self.connect(table)
        result = []
        try:
            result = db.fetch(query).items
        finally:
            yield result
            db.client.close()

    @contextlib.asynccontextmanager
    async def First(self, table, query=None):
        query = query or {}
        db = await self.connect(table)
        result = None
        try:
            result = db.fetch(query).items
        finally:
            yield result
            db.client.close()

    @contextlib.asynccontextmanager
    async def Last(self, table, query={}):
        db = await self.connect(table)
        result = None
        try:
            result =  db.fetch(query).last
        finally:
            yield result
            db.client.close()
            
    @contextlib.asynccontextmanager
    async def Put(self, table, data):
        db = await self.connect(table)
        result = None
        try:
            result = db.put(data)
        finally:
            yield result
            db.client.close()
            
    @contextlib.asynccontextmanager
    async def Insert(self, table, data):
        key = data.get("key")
        if not key:
            raise AttributeError("a key is necessary")
        db = await self.connect(table)
        result = None
        try:
            result = db.insert(data)
        finally:
            yield result
            db.client.close()

    @contextlib.asynccontextmanager
    async def Delete(self, table, key):
        db = await self.connect(table)
        result = None
        try:
            yield db.delete(key=key)
            result = True
        finally:
            yield result
            db.client.close()

    @contextlib.asynccontextmanager
    async def GetOrCreate(self, table: str, data: Dict[ str, Any ]):
        code = data.get('meta', {}).get('code', None) 
        assert code != None, 'CODE could not be found'
        exist, created = None, None
        base = await self.connect(table)
        result = base.fetch({'meta.code': code}).items
        if len(result) >= 1:
            exist = result[0]
        else:
            exist = None
            result = base.put(utils.json_encode(data))
            if result:
                created = result
        try:
            yield exist, created
        finally:
            base.client.close()
            
    @contextlib.asynccontextmanager
    async def CheckCode(self, table, code):
        db = await self.connect(table)
        result = None
        try:
            result = db.fetch({'meta.code': code}).items[0]
        finally:
            yield result
            db.client.close()
            
    @contextlib.asynccontextmanager
    async def SearchPersonByName(self, table, name):
        db = await self.connect(table)
        result = []
        try:
            result = db.fetch({'fullname?contains': name}).items
        finally:
            yield result
            db.client.close()
    
    @contextlib.asynccontextmanager
    async def Update(self, table: str, updates: dict, key: str):
        db = await self.connect(table)
        update_dict = {}
        for k, v in updates.items():
            if k == 'set':
                for k1, v1 in v.items():
                    update_dict[k1] = v1
            elif k == 'increment':
                for k1, v1 in v.items():
                    update_dict[k1] = db.util.increment(v1)
            elif k == 'append':
                for k1, v1 in v.items():
                    update_dict[k1] = db.util.append(v1)
            elif k == 'prepend':
                for k1, v1 in v.items():
                    update_dict[k1] = db.util.prepend(v1)
            elif k == 'trim':
                for k1 in v:
                    update_dict[k1] = db.util.trim()
        result = None
        try:
            result = db.update(updates, key)
            yield result
        except BaseException as e:
            print(e.__dict__)
            yield e
        finally:
            db.client.close()

            
class Database(AbstractDB):
    
    def __init__(self, key, project=Project.BASE):
        super().__init__(key=key, project=project)

        
class Connection(ABC):
    def __init__(self, 
                 base = Database(settings.config.get('BASE_PROJECT', cast=Secret), project=Project.BASE), 
                 log = Database(settings.config.get('LOG_PROJECT', cast=Secret), project=Project.LOG),
                 client = Database(settings.config.get('CLIENT_PROJECT', cast=Secret), project=Project.CLIENT),
                 user = Database(settings.config.get('USER_PROJECT', cast=Secret), project=Project.USER), 
                 ) -> None:
        self.base = base
        self.client = client
        self.log = log
        self.user = user
        


#!/usr/bin/env python
# coding: utf-8

__all__ = [
    'BaseElement',
    'Element',
    'Anchor',
    'Header',
    'Div',
    'OL',
    'Ul',
    'FormField',
    'Fieldset',
    'Input',
    'Select',
    'TextArea',
    'Form',
    'Li',
    'Option',
    'DataList'
    ]

from abc import ABC 
from enum import Enum
from dataclasses import dataclass, field, MISSING
from typing import Optional, Union, Any, List, Dict, Tuple, TypeVar
from markupsafe import Markup

from metadoctor.enumeration import Variable 


is_defined = Variable.is_defined
exist = Variable.exist


HTML5 = '<!doctype html>'
HTML_LANG = '<html lang="pt">'
META_CHARSET = '<meta charset="utf-8">'
META_VIEWPORT = '<meta name="viewport" content="width=device-width, initial-scale=1">'
BOOTSTRAP5_CSS = '<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC" crossorigin="anonymous">'
BOOTSTRAP5_JS = '<script src="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/js/bootstrap.bundle.min.js" integrity="sha384-MrcW6ZMFYlzcLA8Nl+NtUVF0sA7MsXsP1UyJoMp4YLEuNSfAP+JcXn/tWtIaxVXM" crossorigin="anonymous"></script>'
INPUT_STYLE = 'width: 100%; padding: 12px 20px; margin: 8px 0; display: inline-block; border: 1px solid #ccc; border-radius: 4px; box-sizing: border-box;'
POSITIONAL_HTML_ELEMENT_ATTRIBUTES = 'required hidden disabled multiple readonly'.split()
KEYWORD_VALUE_HTML_ELEMENT_ATTRIBUTES = 'placeholder max min minlength maxlength step pattern'.split()



class BaseEnum(Enum):
    def __str__(self):
        return self.value
    
    def __repr__(self):
        return f'{type(self).__name__}.{self.name}'
    
    @classmethod
    def members(cls):
        return [m for m in cls.__members__.values()]
    
    
@dataclass 
class BaseElement(ABC):
    class TOF(BaseEnum):
        TRUE = 'true'
        FALSE = 'false'
    class YON(BaseEnum):
        YES = 'yes'
        NO = 'no'
    tag: str 
    pk: Optional[str] = None
    klass: Optional[str] = None
    style: Optional[str] = None
    title: Optional[str] = None
    accesskey: Optional[str] = None
    lang: Optional[str] = None
    hidden: Optional[bool] = None
    tabindex: Optional[int] = None
    contenteditable: Optional[TOF] = None
    spellcheck: Optional[TOF] = None
    draggable: Optional[TOF] = None
    translate: Optional[YON] = None
    data_attr: Optional[Tuple[str, str]] = None
    content: List[Union[str, 'BaseElement', None]] = field(default_factory=list)
    is_empty: bool = False 
    has_items: bool = False 
    has_options: bool = False 
    # Clipboard Events  
    oncopy: Optional[str] = None # Fires when the user copies the content of an element
    oncut: Optional[str] = None # Fires when the user cuts the content of an element
    onpaste: Optional[str] = None # Fires when the user pastes some content in an element
    # Drag Events
    ondrag: Optional[str] = None # Script to be run when an element is dragged
    ondragend: Optional[str] = None # Script to be run at the end of a drag operation
    ondragenter: Optional[str] = None # Script to be run when an element has been dragged to a valid drop target
    ondragleave: Optional[str] = None # Script to be run when an element leaves a valid drop target
    ondragover: Optional[str] = None # Script to be run when an element is being dragged over a valid drop target
    ondragstart: Optional[str] = None # Script to be run at the start of a drag operation
    ondrop: Optional[str] = None # Script to be run when dragged element is being dropped
    onscroll: Optional[str] = None # Script to be run when an element's scrollbar is being scrolled
        
    
        
    def __add__(self, other):
        assert isinstance(other, (str, BaseElement, Markup)), f'{type(self).__name__} can be added to str, BaseElemente or Markup instances only'
        other = other if isinstance(other, (str, Markup)) else str(other)
        return self.__html__() + other
    
    @property
    def _tag(self):
        return self.tag.lower()
    
    @property
    def _pk(self):
        return f'id="{self.pk}"' if self.pk else None 
    
    @property
    def _data_attr(self):
        return f'data-{self.data_attr[0]}="{self.data_attr[1]}"' if self.data_attr else None
                         
    @property        
    def _klass(self):
        return f'class="{self.klass}"' if self.klass else None 
            
    @property        
    def _style(self):
        return f'style="{self.style}"' if self.style else None 
                         
    @property        
    def _title(self):
        return f'title="{self.title}"' if self.title else None  
                         
    @property        
    def _accesskey(self):
        return f'accesskey="{self.accesskey}"' if self.accesskey else None 

    @property        
    def _tabindex(self):
        return f'tabindex="{self.tabindex}"' if self.tabindex else None 
                         
    @property        
    def _lang(self):
        if type(self) == HTML:
            return f'lang="{self.lang}"' if self.lang else f'lang="pt"' 
        return f'lang="{self.lang}"' if self.lang else None 
                         
    @property        
    def _hidden(self):
        return 'hidden' if self.hidden in [True, self.YON.YES, 'yes', self.TOF.TRUE, 'true'] else None  
                       
    @property        
    def _spellcheck(self):
        return 'spellcheck="true"' if self.spellcheck in [True, self.YON.YES, 'yes', self.TOF.TRUE, 'true'] else None  
    
    
    @property        
    def _draggable(self):
        return 'draggable="true"' if self.draggable in [True, self.YON.YES, 'yes', self.TOF.TRUE, 'true'] else None  

    @property        
    def _translate(self):
        return 'translate="yes"' if self.translate in [True, self.YON.YES, 'yes', self.TOF.TRUE, 'true'] else None  
    
    
    @property        
    def _contenteditable(self):
        return f'contenteditable="true"' if self.contenteditable in [True, self.YON.YES, 'yes', self.TOF.TRUE, 'true'] else None 
    
    def _global_attributes_str(self) -> str:
        return ' '.join(filter(lambda x: x != None, [
            self._pk, 
            self._klass, 
            self._style, 
            self._title, 
            self._accesskey, 
            self._tabindex, 
            self._lang, 
            self._hidden, 
            self._contenteditable,
            self._translate,
            self._spellcheck,
            self._draggable
        ]))
    
#     @classmethod
#     def parse_item(cls, item):
#         return str(Li(item))
        

    def compile_items(self):
        def parse_item(item):
            return Li(str(item))
        result = []
        if isinstance(self.content, (tuple, list)):
            for item in self.content:
                result.append(parse_item(item))
        elif isinstance(self.content, str):
            result.append(parse_item(item))
        return result
    
    
    def compile_options(self):
        def parse_option(value, option, selected=None, label=None):
            return Option(value=value, content=option, selected=None, label=label) 
        result = []
        if isinstance(self.content, (tuple, list)):
            for item in self.content:
                if isinstance(item, BaseEnum):
                    result.append(parse_option(item.name, item.value))
                elif isinstance(item, (tuple, list)):
                    result.append(parse_option(item[0], item[1]))
                elif isinstance(item, str):
                    result.append(parse_option(item, item))
        return result 
    
    
    def compile_text(self):
        if isinstance(self.content, str):
            return self.content 
        elif isinstance(self.content, (tuple, list)):
            return ''.join([str(x) for x in self.content])
            
    
    @property        
    def _content(self) -> str:
        if self.has_options:
            return ''.join([str(x) for x in self.compile_options()])
        elif self.has_items:
            return ''.join([str(x) for x in self.compile_items()])
        else:
            return self.compile_text()
#         result = []
#         if self.has_items:
#             for item in self.content:
#                 result.append(self.parse_item(str(item)))
#         elif self.has_options:
#             if isinstance(self.content, str):
#                 result.append(self.content)
#             elif isinstance(self.content, list):
#                 for item in self.content:
#                     result.append(
#                         self.parse_option(
#                             *item if isinstance(item, (tuple, list)) 
#                             else (item, item) if isinstance(item, str) 
#                             else (item.name, item.value) if isinstance(item, Enum) 
#                             else (str(item), str(item))))
#         elif self.is_empty:
#             return ''
#         else:
#             if isinstance(self.content, str):
#                 result.append(self.content)
#             elif isinstance(self.content, (tuple, list)):
#                 for item in self.content:
#                     result.append(str(item))   
#         return ''.join(result)
    
                         
    def _local_attributes_str(self) -> str:
        return '' 

    def _attributes(self) -> str:
        return ' '.join([x for x in [self._global_attributes_str(), self._local_attributes_str()] if len(x) > 0])
        
    def _repr_html_(self):
        if self.is_empty:
            return f'<{" ".join([self._tag, self._attributes()])}>'
        return f'<{" ".join([self._tag, self._attributes()])}>{self._content}</{self._tag}>'
    
    def __str__(self):
        return self._repr_html_()  
    
    def __html__(self):
        return Markup(self._repr_html_())  
    

@dataclass 
class HTML(BaseElement):
    tag: str = 'HTML' 
    
        
    def __repr_html__(self):
        text = HTML5
        text += f'<html lang="{self.lang or "pt"}">'


@dataclass
class Body(BaseElement):
    onafterprint: Optional[str] = None # Script to be run after the document is printed
    onbeforeprint: Optional[str] = None # Script to be run before the document is printed
    onbeforeunload: Optional[str] = None # Script to be run when the document is about to be unloaded
    onerror: Optional[str] = None # Script to be run when an error occurs
    onhashchange: Optional[str] = None # Script to be run when there has been changes to the anchor part of the a URL
    onload: Optional[str] = None # Fires after the page is finished loading
    onmessage: Optional[str] = None # Script to be run when the message is triggered
    onoffline: Optional[str] = None # Script to be run when the browser starts to work offline
    ononline: Optional[str] = None # Script to be run when the browser starts to work online
    onepagehide: Optional[str] = None # Script to be run when a user navigates away from a page
    onpageshow: Optional[str] = None # Script to be run when a user navigates to a page
    onpopstate: Optional[str] = None # Script to be run when the window's history changes
    onresize: Optional[str] = None # Fires when the browser window is resized
    onstorage: Optional[str] = None # Script to be run when a Web Storage area is updated
    onunload: Optional[str] = None # Fires once a page has unloaded (or the browser window has been closed)
        
    def _local_attributes_str(self) -> str:
        result = []
        if self.onafterprint: result.append(f'onafterprint="{self.onafterprint}"')
        if self.onbeforeprint: result.append(f'onbeforeprint="{self.onbeforeprint}"')
        if self.onbeforeunload: result.append(f'onbeforeunload="{self.onbeforeunload}"')
        if self.onerror: result.append(f'onerror="{self.onerror}"')
        if self.onload: result.append(f'onload="{self.onload}"')
        if self.onmessage: result.append(f'onmessage="{self.onmessage}"')
        if self.onoffline: result.append(f'onoffline="{self.onoffline}"')
        if self.ononline: result.append(f'ononline="{self.ononline}"')
        if self.onepagehide: result.append(f'onepagehide="{self.onepagehide}"')
        if self.onpageshow: result.append(f'onpageshow="{self.onpageshow}"')
        if self.onpopstate: result.append(f'onpopstate="{self.onpopstate}"')
        if self.onresize: result.append(f'onresize="{self.onresize}"')
        if self.onstorage: result.append(f'onstorage="{self.onstorage}"')
        if self.onunload: result.append(f'onunload="{self.onunload}"')            
        return ' '.join(result)
    

@dataclass 
class Element(BaseElement):
    pass 


@dataclass 
class Form(BaseElement):
    tag: str = 'FORM'
    action: str = '.'
    # events 
    onblur: Optional[str] = None # Fires the moment that the element loses focus
    onchange: Optional[str] = None # Fires the moment when the value of the element is changed
    oncontextmenu: Optional[str] = None # Script to be run when a context menu is triggered
    onfocus: Optional[str] = None # Fires the moment when the element gets focus
    oninput: Optional[str] = None # Script to be run when an element gets user input
    oninvalid: Optional[str] = None # Script to be run when an element is invalid
    onreset: Optional[str] = None # Fires when the Reset button in a form is clicked
    onsearch: Optional[str] = None # Fires when the user writes something in a search field (for <input="search">)
    onselect: Optional[str] = None # Fires after some text has been selected in an element
    onsubmit: Optional[str] = None # Fires when a form is submitted
    onblur: Optional[str] = None # 
        
    class Target(BaseEnum):
        BLANK = '_blank' # The response is displayed in a new window or tab
        TOP = '_top' # The response is displayed in the full body of the window
        SELF = '_self' # The response is displayed in the current window
        PARENT = '_parent' # The response is displayed in the parent frame
        FRAMENAME = 'framaneme' # The response is displayed in a named iframe
        
    target: Optional[Target] = None
        
    class Method(BaseEnum):
        GET = 'get'
        POST = 'post'
        
    method: Method = Method.GET
        
    class Autocomplete(BaseEnum):
        ON = 'on'
        OFF = 'off'
        
    autocomplete: Optional[Autocomplete] = None
    novalidate: bool = False
        
    class Rel(BaseEnum):
        EXTERNAL = 'external' # Specifies that the referenced document is not a part of the current site
        HELP = 'help' # Links to a help document
        LICENCE = 'licence' # Indicates that the main content of the current document is covered by the copyright license described by the referenced document.
        NEXT = 'next' # Indicates that the current document is a part of a series and that the next document in the series is the referenced document.
        NOFOLLOW = 'nofollow' # Links to an unendorsed document, like a paid link. Indicates that the current document's original author or publisher does not endorse the referenced document.
        NOOPENER = 'noopener' # Creates a top-level browsing context that is not an auxiliary browsing context if the hyperlink would create either of those, 
                              # to begin with (i.e., has an appropriatetargetattribute value).
        NOREFERRER = 'noreferrer' # Specifies that the browser should not send a HTTP referrer header if the user follows the hyperlink. No Referer header will be included. 
                                  # Additionally, has the same effect as noopener.
        OPENER = 'opener' # Creates an auxiliary browsing context if the hyperlink would otherwise create a top-level browsing context that 
                          # is not an auxiliary browsing context (i.e., has "_blank" as target attribute value).
        PREV = 'prev' # Indicates that the current document is a part of a series and that the previous document in the series is the referenced document.
        SEARCH = 'search' # Gives a link to a resource that can be used to search through the current document and its related pages.
    
    rel: Optional[Rel] = None
    
    def _local_attributes_str(self):
        data = [
            f'action="{self.action}" method="{self.method}"',
            f'autocomplete="{self.autocomplete}"' if self.autocomplete else None,
            f'rel="{self.rel}"' if self.rel else None,
            f'target="{self.target}"' if self.target else None,
            f'novalidate' if self.novalidate else None,
            ]
        return ' '.join(filter(lambda x: x != None, data))
    
    @property
    def _content(self):
        text = super()._content 
        text += str(Input(input_type=Input.Type('submit')))
        return text


@dataclass 
class Anchor(Element):
    tag: str = 'A' 
    href: str =  field(default=...)

    def _local_attributes_str(self):
        return f'href="{self.href}"'
    
    @property
    def _content(self):
        return self.content if isinstance(self.content, str) else ''.join(self.content)
    

@dataclass 
class Header(Element):
    tag: int
    
    @property
    def _tag(self):
        assert isinstance(self.tag, int)
        assert self.tag > 0
        assert self.tag < 7
        return 'h' + str(self.tag)
    

@dataclass 
class Fieldset(Element):
    tag: str = 'FIELDSET'
    legend: Optional[str] = None
        
    @property
    def _klass(self):
        klass = super()._klass
        return f'class="{klass} row"' if klass else 'class="row"'
    
    @property
    def _content(self):
        if isinstance(self.content, str):
            return self.content
        elif isinstance(self.content, (tuple, list)):
            return ''.join([str(x) for x in self.content])
        return super()._content
    
    def _repr_html_(self):
        legend = None
        if self.legend:
            legend = f'<legend >{self.legend} </legend>'
        if legend:
            fieldset = f'<{" ".join([self._tag, self._attributes()])}>{legend}{self._content}</{self._tag}>'
        else:
            fieldset = f'<{" ".join([self._tag, self._attributes()])}>{self._content}</{self._tag}>'
        return fieldset
    

@dataclass 
class Li(Element):
    tag: str = 'LI'

    
@dataclass 
class Option(Element):
    tag: str = 'OPTION'
    value: Optional[str] = None
    label: Optional[str] = None
    selected: bool = False
    optgroup: Optional[str] = None
        
    @property 
    def _content(self):
        return self.content.value if isinstance(self.content, BaseEnum) else str(self.content)
        
        
    def _attributes(self):
        value = self.value or self._content
        value = f'value="{value}"'
        if self.selected: value += ' selected'
        return value 


@dataclass 
class Ol(Element):
    tag: str = 'OL'
    has_items: bool = True


@dataclass 
class Ul(Element):
    tag: str = 'UL'
    has_items: bool = True
    
    
@dataclass 
class Div(Element):
    tag: str = 'DIV'
        


@dataclass
class FormField(Element):
    label: Optional[str] = None
    required: Optional[bool] = None
    readonly: Optional[bool] = None
    disabled: Optional[bool] = None
    hidden: Optional[bool] = None
    max: Optional[bool] = None
    min: Optional[bool] = None
    minlength: Optional[bool] = None
    maxlength: Optional[bool] = None
    step: Optional[bool] = None
    pattern: Optional[bool] = None
    placeholder: Optional[str] = None
    name: Optional[str] = None
    value: Optional[str] = None


    class Type(BaseEnum):
        TEXT='text'
        DATETIME_LOCAL='datetime-local'
        DATE='date'
        SUBMIT='submit'
        NUMBER='number'
        WEEK='week'
        COLOR='color'
        IMAGE='image'
        MONTH='month'
        RANGE='range'
        BUTTON="button"
        CHECKBOX="checkbox"
        EMAIL="email"
        FILE="file"
        HIDDEN="hidden"
        PASSWORD="password"
        RADIO="radio"
        RESET="reset"
        SEARCH="search"
        TEL="tel"
        TIME="time"
        URL="url"

        def __str__(self):
            return f'type="{self.value}"'
        
    input_type: Optional[Type] = None
    
    @property        
    def _klass(self):
        return f'class="{self.klass} form-control form-field"' if self.klass else 'class="form-control form-field"' 
    
    
    def _local_attributes_str(self):
        result = []
        if self.input_type:
            result.append(f'type="{self.input_type}"')
        if self.required:
            result.append('required')
        if self.readonly:
            result.append('readonly')
        if self.disabled:
            result.append('disabled')
        if self.hidden:
            result.append('hidden')
        if self.max:
            result.append(f'max="{self.max}"')
        if self.min:
            result.append(f'min="{self.min}"')
        if self.minlength:
            result.append(f'minlength="{self.minlength}"')
        if self.maxlength:
            result.append(f'maxlength="{self.maxlength}"')
        if self.step:
            result.append(f'step="{self.step}"')
        if self.pattern:
            result.append(f'pattern="{self.pattern}"')
        if self.value and not type(self) == Select:
            result.append(f'value="{self.value}"')
        return ' '.join(result) 
            
    
    @property
    def _pk(self):
        if self.pk:
            return f'id="{self.pk}" name="{self.name or self.pk}"'
        return
    
    def _repr_html_(self):
        field_wraper = lambda: f'<{" ".join([self._tag, self._attributes()])}>{self._content}</{self._tag}>'
        html = lambda: f'<label for="{self.pk}" class="form-label">{self.label or self.pk} {field_wraper()}</label>'.replace('</input>','')
        return html()
        
#         label = None
#         if self.label:
#             if self.pk:
#                 label = f'<label for="{self.pk}" class="form-label">{self.label} </label>'
#             else:
#                 label = f'<label class="form-label">{self.label} </label>'
#         if self.is_empty:
#             if label:
#                 htmlfield = f'{label}<{" ".join([self._tag, self._attributes()])}>'
#             else:
#                 htmlfield = f'<{" ".join([self._tag, self._attributes()])}>'
#         else:
#             if label:
#                 htmlfield = f'{label}<{" ".join([self._tag, self._attributes()])}>{self._content}</{self._tag}>'
#             else:
#                 htmlfield = f'<{" ".join([self._tag, self._attributes()])}>{self._content}</{self._tag}>'
#         return htmlfield
    

@dataclass 
class Select(FormField):
    tag: str = 'SELECT'
    has_options: bool = True
    input_type: Optional[str] = None
    form: Optional[str] = None # the form id
    multiple: Optional[str] = None
    size: Optional[int] = None
        
    def _attributes(self):
        result = [super()._attributes()]
        if self.form:
            result.append(f'form="{self.form}"')
        if self.multiple:
            result.append(self.multiple)
        if self.size:
            result.append(f'size="{self.size}"')
        
        return ' '.join(result)
    
    
@dataclass 
class TextArea(FormField):
    tag: str = 'TEXTAREA'
    
    
@dataclass 
class Input(FormField):
    tag: str = 'INPUT'
    is_empty: str = True
    list: Optional[str] = None
    
    def __post_init__(self):
        if self.input_type:
            self.input_type = self.Type(self.input_type)
    
    def _local_attributes_str(self):
        return ' '.join(filter(
            lambda x: x != None,
            [
                str(self.Type('text')) if not self.input_type else str(self.input_type),
                f'list="{self.list}"' if self.list else None
            ]
        ))
    

    
    
@dataclass 
class DataList(FormField):
    tag: str = 'INPUT'
    has_options: bool = True
    input_type: Optional[str] = None
    
    @property
    def list_name(self):
        return f'{self.pk}List'
    
    def _local_attributes_str(self):
        text = super()._local_attributes_str()
        text += f' list="{self.list_name}"'
        return text 

    
    @property        
    def _content(self) -> str:
        datalist = lambda options: f'<datalist id="{self.list_name}">{options}</datalist>'
        return datalist(super()._content)
    

    
        
@dataclass
class Media(Element):
    # events 
    onabort: Optional[str] = None # 
    oncanplay: Optional[str] = None # Script to be run when a file is ready to start playing (when it has buffered enough to begin)
    oncanplaythrough: Optional[str] = None # Script to be run when a file can be played all the way to the end without pausing for buffering
    oncuechange: Optional[str] = None # Script to be run when the cue changes in a <track> element
    ondurationchange: Optional[str] = None # Script to be run when the length of the media changes
    onemptied: Optional[str] = None # Script to be run when something bad happens and the file is suddenly unavailable (like unexpectedly disconnects)
    onended: Optional[str] = None # Script to be run when the media has reach the end (a useful event for messages like "thanks for listening")
    onerror: Optional[str] = None # Script to be run when an error occurs when the file is being loaded
    onloadeddata: Optional[str] = None # Script to be run when media data is loaded
    onloadedmetadata: Optional[str] = None # Script to be run when meta data (like dimensions and duration) are loaded
    onloadstart: Optional[str] = None # Script to be run just as the file begins to load before anything is actually loaded
    onpause: Optional[str] = None # Script to be run when the media is paused either by the user or programmatically
    onplay: Optional[str] = None # Script to be run when the media is ready to start playing
    onplaying: Optional[str] = None # Script to be run when the media actually has started playing
    onprogress: Optional[str] = None # Script to be run when the browser is in the process of getting the media data
    onratechange: Optional[str] = None # Script to be run each time the playback rate changes (like when a user switches to a slow motion or fast forward mode)
    onseeked: Optional[str] = None # Script to be run when the seeking attribute is set to false indicating that seeking has ended
    onseeking: Optional[str] = None # Script to be run when the seeking attribute is set to true indicating that seeking is active
    onstalled: Optional[str] = None # Script to be run when the browser is unable to fetch the media data for whatever reason
    onsuspend: Optional[str] = None # Script to be run when fetching the media data is stopped before it is completely loaded for whatever reason
    ontimeupdate: Optional[str] = None # Script to be run when the playing position has changed (like when the user fast forwards to a different point in the media)
    onvolumechange: Optional[str] = None # Script to be run each time the volume is changed which (includes setting the volume to "mute")
    onwaiting: Optional[str] = None # Script to be run when the media has paused but is expected to resume (like when the media pauses to buffer more data)
        
        
    def _attributes(self):
        result = [super()._attributes()]
        if self.onabort: 
            result.append(f'onabort="{self.onboard}"')
        if self.oncanplay: 
            result.append(f'oncanplay="{self.oncanplay}"')
        if self.oncanplaythrough: 
            result.append(f'oncanplaythrough="{self.oncanplaythrough}"')
        if self.ondurationchange: 
            result.append(f'ondurationchange="{self.ondurationchange}"')
        if self.onerror: 
            result.append(f'onerror="{self.onerror}"')
        if self.onloadeddata: 
            result.append(f'onloadeddata="{self.onloadeddata}"')
        if self.onloadedmetadata: 
            result.append(f'onloadedmetadata="{self.onloadedmetadata}"')
        if self.onloadstart: 
            result.append(f'onloadstart="{self.onloadstart}"')
        if self.onpause: 
            result.append(f'onpause="{self.onpause}"')
        if self.onplay: 
            result.append(f'onplay="{self.onplay}"')
        if self.onplaying: 
            result.append(f'onplaying="{self.onplaying}"')
        if self.onprogress: 
            result.append(f'onprogress="{self.onprogress}"')
        if self.onratechange: 
            result.append(f'onratechange="{self.onratechange}"')
        if self.onseeked: 
            result.append(f'onseeked="{self.onseeked}"')
        if self.onseeking: 
            result.append(f'onseeking="{self.onseeking}"')
        if self.onstalled: 
            result.append(f'onstalled="{self.onstalled}"')
        if self.onsuspend: 
            result.append(f'onsuspend="{self.onsuspend}"')
        if self.ontimeupdate: 
            result.append(f'ontimeupdate="{self.ontimeupdate}"')
        if self.onvolumechange: 
            result.append(f'onvolumechange="{self.onvolumechange}"')
        if self.onwaiting: 
            result.append(f'onwaiting="{self.onwaiting}"')
        return ' '.join(result)
            

@dataclass
class Img(Media):
    tag: str = 'IMG'
        

@dataclass
class Audio(Media):
    tag: str = 'AUDIO'


@dataclass
class Embed(Media):
    tag: str = 'IMG'
        
        
@dataclass
class Object(Media):
    tag: str = 'OBJECT'

        
@dataclass
class Video(Media):
    tag: str = 'VIDEO'



__version__ = '0.1.6'

from metadoctor import settings
from metadoctor import database
from metadoctor import enumeration
from metadoctor import base 
from metadoctor import utils 
from metadoctor import element
